/* -*-Java-*-
 * ###################################################################
 *
 *  FILE: "Arbre.java"
 *                                    created: 2002-10-29 11:06:56
 *                                last update: 2008-02-22 17:06:06
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          3 rue Joliot-Curie, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/
 *
 *  Description:
 *  Arbre binaire.
 *
 *  spc ARBRE
 *    tends ELEMENT, BOOL
 *    sorte Arbre
 *    oprations
 *      arbreVide  :                       -> Arbre
 *      construire : Arbre, Element, Arbre -> Arbre
 *      racine     : Arbre                 -> Element
 *      filsGauche : Arbre                 -> Arbre
 *      filsDroit  : Arbre                 -> Arbre
 *      estVide    : Arbre                 -> Bool
 *    prconditions a:Arbre
 *      pr racine(a) = .not. estVide(a)
 *    axiomes g,d:Arbre; e:Element
 *      estVide(arbreVide) = vrai
 *      estVide(construire(g,e,d)) = faux
 *      racine(construire(g,e,d)) = e
 *      filsGauche(arbreVide) = arbreVide
 *      filsGauche(construire(g,e,d)) = g
 *      filsDroit(arbreVide) = arbreVide
 *      filsDroit(construire(g,e,d)) = d
 *  fspc
 *
 *  History
 *
 *  modified   by  rev reason
 *  ---------- --- --- -----------
 *  2002-10-29 FBO 1.0 original
 * ###################################################################
 */

package collections;

/*+ArbreBinaire/arbre.java*/
public class Arbre<T> {
  private Noeud racine_;

  // Mthodes correspondant  la spcification algbrique
  /** Construit un arbre vide. */
  public Arbre() {
    racine_ = null;
  }

  /**
   * Construit un arbre dont la racine contient <code>e</code>, a pour
   * fils gauche la racine de l'arbre <code>g</code> et pour fils droit
   * la racine de l'arbre <code>d</code>. Les arbres <code>g</code> et
   * <code>d</code> sont vides aprs l'opration.
   */
  public Arbre(Arbre<T> g, T e, Arbre<T> d) {
    racine_ = new Noeud(g.racine_, e, d.racine_);
    g.racine_ = null; // car on a rcupr ses noeuds
    d.racine_ = null; // car on a rcupr ses noeuds
  }

  /**
   * Rend le fils gauche de la racine de cet arbre et supprime de
   * cet arbre les noeuds correspondant.
   */
  public Arbre<T> filsGauche() {
    Arbre<T> fils = new Arbre<T>();
    if (racine_ == null) {
      return fils;
    }
    fils.racine_ = racine_.filsGauche;
    racine_.filsGauche = null; // car on a rcupr ses noeuds
    return fils;
  }

  /**
   * Rend le fils droit de la racine de cet arbre et supprime de
   * cet arbre les noeuds correspondant.
   */
  public Arbre<T> filsDroit() {
    Arbre<T> fils = new Arbre<T>();
    if (racine_ == null) {
      return fils;
    }
    fils.racine_ = racine_.filsDroit;
    racine_.filsDroit = null; // car on a rcupr ses noeuds
    return fils;
  }

  /**
   * Rend l'objet contenu par la racine de cet arbre.
   */
  public T racine() {
    prerequis(racine_ != null);
    return racine_.element;
  }

  /** Indique si cet arbre est vide. */
  public boolean estvide() {
    return racine_ == null;
  }
  protected void prerequis(boolean b) {
    if (!b) {
      throw new Error("Prrequis non valide dans " + getClass().getName());
    }
  }
/*-*/
// Mthodes utilisant des curseurs
/*+ArbreBinaire/methodesAvecCurseurs.java*/
  /** Rend un curseur dsignant la racine de cet arbre. */
  public Curseur<T> curseurRacine() {
    return racine_;
  }

  /**
   * Rend un curseur dsignant le fils gauche du noeud dsign
   * par <code>c</code>.
   */
  public Curseur<T> versFilsGauche(Curseur<T> c) {
    return ( (Noeud) c).filsGauche;
  }

  /**
   * Rend un curseur dsignant le fils droit du noeud dsign
   * par <code>c</code>.
   */
  public Curseur<T> versFilsDroit(Curseur<T> c) {
    return ( (Noeud) c).filsDroit;
  }

  /** Rend l'objet contenu dans le noeud dsign par <code>c</code>. */
  public T element(Curseur<T> c) {
    return ( (Noeud) c).element;
  }

  /**
   * Remplace le contenu du noeud dsign par <code>c</code> par
   * l'objet <code>e</code>.
   */
  public void remplacer(T e, Curseur<T> c) {
    ( (Noeud) c).element = e;
  }

  /**
   * Ajoute un noeud contenant <code>e</code> en tant que
   * fils gauche du noeud dsign par <code>c</code>.
   * choue si ce noeud a dj un fils gauche.
   */
  public void ajouterFilsGauche(T e, Curseur<T> c) {
    Noeud n = (Noeud) c;

    if (n.filsGauche != null) {
      throw new Error("Le noeud a dj un fils gauche");
    }
    n.filsGauche = new Noeud(e);
  }

  /**
   * Ajoute les noeuds de l'arbre <code>a</code>  cet arbre
   * de faon  ce que la racine de <code>a</code> soit le
   * fils gauche du noeud dsign par <code>c</code>. L'arbre
   * <code>a</code> est vide aprs l'opration.
   * choue si le noeud a dj un fils gauche.
   */
  public void ajouterFilsGauche(Arbre<T> a, Curseur<T> c) {
    Noeud n = (Noeud) c;

    if (n.filsGauche != null) {
      throw new Error("Le noeud a dj un fils gauche");
    }
    n.filsGauche = a.racine_;
    a.racine_ = null; // car on a rcupr ses noeuds
  }

  /**
   * Ajoute un noeud contenant <code>e</code> en tant que
   * fils droit du noeud dsign par <code>c</code>.
   * choue si ce noeud a dj un fils droit.
   */
  public void ajouterFilsDroit(T e, Curseur<T> c) {
    Noeud n = (Noeud) c;

    if (n.filsDroit != null) {
      throw new Error("Le noeud a dj un fils droit");
    }
    n.filsDroit = new Noeud(e);
  }

  /**
   * Ajoute les noeuds de l'arbre <code>a</code>  cet arbre
   * de faon  ce que la racine de <code>a</code> soit le
   * fils droit du noeud dsign par <code>c</code>. L'arbre
   * <code>a</code> est vide aprs l'opration.
   * choue si le noeud a dj un fils droit.
   */
  public void ajouterFilsDroit(Arbre<T> a, Curseur<T> c) {
    Noeud n = (Noeud) c;

    if (n.filsDroit != null) {
      throw new Error("Le noeud a dj un fils droit");
    }
    n.filsDroit = a.racine_;
    a.racine_ = null; // car on a rcupr ses noeuds
  }

  /**
   * Supprime le fils gauche (et ses descendants) du noeud
   * dsign par <code>c</code>.
   */
  public void supprimerFilsGauche(Curseur<T> c) {
    ( (Noeud) c).filsGauche = null;
  }

  /**
   * Supprime le fils droit (et ses descendants) du noeud
   * dsign par <code>c</code>.
   */
  public void supprimerFilsDroit(Curseur<T> c) {
    ( (Noeud) c).filsDroit = null;
  }
/*-*/
  public boolean memePosition(Curseur<T> c1, Curseur<T> c2) {
    return c1 == c2;
  }

  public Arbre<T> dupliquer() {
    Arbre<T> res = new Arbre<T>();
    if (racine_ != null) {
      res.racine_ = racine_.dupliquer(res);
    }
    return res;
  }

  public boolean equals(Arbre<T> a) {
    if (racine_ == null && a.racine_ == null) {
      return true;
    }
    if (racine_ == null && a.racine_ != null) {
      return false;
    }
    if (racine_ != null && a.racine_ == null) {
      return false;
    }
    return racine_.memeArbre(a.racine_);
  }

  /**
   * Reprsente un noeud de l'arbre, correspond  la notion de curseur.
   */
/*+ArbreBinaire/noeud.java*/
  private class Noeud extends Curseur<T> {
    // Les membres de cette classe sont publics car Noeud est
    // une classe prive de Arbre et n'est donc vue par aucune autre
    // classe. Comme elle joue un rle privilgi dans l'implmentation
    // de Arbre, on laisse Arbre accder directement  ses membres, ce
    // qui rend le code plus performant (moins d'invocations de mthodes)
    // et plus simple.
    // Cette pratique n'est justifie que dans de rares cas particuliers
    // comme celui-ci. En gnral, les membres d'une classe doivent tre
    // privs ou protgs, et leur valeur ne peut tre consulte ou modifie
    // que par invocation de mthodes. Mme si cela peut sembler lourd 
    // premire vue, cela rend la classe beaucoup plus utile en tant
    // que classe de base, et dcouple les choix d'implmentation de
    // l'interface publique.

    /** Objet contenu dans ce noeud. */
    public T element;

    /** Fils gauche de ce noeud. */
    public Noeud filsGauche;

    /** Fils droit de ce noeud. */
    public Noeud filsDroit;

    /** Construit un noeud contenant l'objet <code>e</code>. */
    public Noeud(T e) {
      this(null, e, null);
    }

    /**
     * Construit un noeud contenant l'objet <code>e</code>, ayant
     * pour fils gauche <code>fg</code> et pour fils droit <code>fd</code>.
     */
    public Noeud(Noeud fg, T e, Noeud fd) {
      element = e;
      filsGauche = fg;
      filsDroit = fd;
    }

/*P+*/
    public Noeud dupliquer(Arbre<T> proprietaire) {
      Noeud fg = null, fd = null;
      if (filsGauche != null) {
        fg = filsGauche.dupliquer(proprietaire);
      }
      if (filsDroit != null) {
        fd = filsDroit.dupliquer(proprietaire);
      }
      return new Noeud(fg, element, fd);
    }

    public boolean memeArbre(Noeud n) {
      if (n == null) return false;
      if (element != n.element) return false;
      if (filsGauche == null && n.filsGauche != null) return false;
      if (filsGauche != null && !filsGauche.memeArbre(n.filsGauche)) return false;
      if (filsDroit == null && n.filsDroit != null) return false;
      if (filsDroit != null && !filsDroit.memeArbre(n.filsDroit)) return false;
      return true;
    }
  }
  /*-*/
}
