/* -*-Java-*-
 * ###################################################################
 *
 *  FILE: "Foret.java"
 *                                    created: 2002-10-29 11:13:32
 *                                last update: 2008-02-22 17:04:39
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          3 rue Joliot-Curie, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/
 *
 *  Description:
 *  Fort, ou arbre planaire.
 *
 *  spc FORET
 *    tends ELEMENT, BOOL
 *    sorte Foret
 *    oprations
 *      foretVide  :                       -> Foret
 *      construire : Foret, Element, Foret -> Foret
 *      racine     : Foret                 -> Element
 *      fille      : Foret                 -> Foret
 *      soeur      : Foret                 -> Foret
 *      estVide    : Foret                 -> Bool
 *    prconditions f:Foret
 *      pr racine(f) = .not. estVide(f)
 *    axiomes f,s:Foret; e:Element
 *      estVide(foretVide) = vrai
 *      estVide(construire(f,e,s)) = faux
 *      racine(construire(f,e,s)) = e
 *      fille(foretVide) = foretVide
 *      fille(construire(f,e,s)) = f
 *      soeur(foretVide) = foretVide
 *      soeur(construire(f,e,s)) = s
 *    fspc
 *
 *  History
 *
 *  modified   by  rev reason
 *  ---------- --- --- -----------
 *  2002-10-29 FBO 1.0 original
 * ###################################################################
 */
package collections;
/*+Foret/foret.java*/
public class Foret<T> {
  private Noeud racine_;

  // Mthodes correspondant  la spcification algbrique

  /** Cre une fort vide. */
  public Foret() {
    racine_ = null;
  }

  /**
   * Cre une fort dont la racine contient <code>e</code>, a pour
   * frres les noeuds de la fort <code>soeur</code>, et pour fils
   * les noeuds de la fort <code>fille</code>.
   */
  public Foret(Foret<T> fille, T e, Foret<T> soeur) {
    racine_ = new Noeud(fille.racine_, e, soeur.racine_);
    fille.racine_ = null; // car on a rcupr ses noeuds
    soeur.racine_ = null; // car on a rcupr ses noeuds
  }

  /**
   * Rend la fort compose des fils de la racine de cette fort.
   * Les noeuds de la fort rendue sont supprims de cette fort.
   */
  public Foret<T> fille() {
    Foret<T> f = new Foret<T>();
    if (racine_ == null) {
      return f;
    }
    f.racine_ = racine_.filsAine;
    racine_.filsAine = null; // car on a rcupr ses noeuds
    return f;
  }

  /**
   * Rend la fort compose des frres de la racine de cette fort.
   * Les noeuds de la fort rendue sont supprims de cette fort.
   */
  public Foret<T> soeur() {
    Foret<T> f = new Foret<T>();
    if (racine_ == null) {
      return f;
    }
    f.racine_ = racine_.frereCadet;
    racine_.frereCadet = null; // car on a rcupr ses noeuds
    return f;
  }

  /**
   * Rend le contenu de la racine de cette fort.
   */
  public T racine() {
    if (racine_ == null) {
      throw new Error("Arbre vide");
    }
    return racine_.contenu;
  }

  /** Indique si cette fort est vide. */
  public boolean estVide() {
    return racine_ == null;
  }

  // Mthodes utilisant des curseurs

  /** Rend un curseur sur la racine de cette fort. */
  public Curseur<T> curseurRacine() {
    return racine_;
  }

  /**
   * Rend un curseur sur le premier fils du noeud dsign
   * par le curseur <code>c</code>.
   */
  public Curseur<T> versFilsAine(Curseur<T> c) {
    return ((Noeud)c).filsAine;
  }

  /**
   * Rend un curseur sur frre cadet (noeud situ immdiatement  droite)
   * du noeud dsign par le curseur <code>c</code>.
   */
  public Curseur<T> versFrereCadet(Curseur<T> c) {
    return ((Noeud)c).frereCadet;
  }

  /**
   * Rend l'lment contenu dans le noeud dsign par <code>c</code>.
   */
  public T element(Curseur<T> c) {
    return ((Noeud)c).contenu;
  }

  /**
   * Remplace le contenu du noeud dsign par <code>c</code>
   * par l'objet <code>e</code>.
   */
  public void remplacer(T e, Curseur<T> c) {
    ((Noeud)c).contenu = e;
  }

  /**
   * Ajoute une noeud contenant <code>e</code> en tant que
   * premier fils (le plus  gauche) du noeud dsign par <code>c</code>.
   */
  public void ajouterFilsAine(T e, Curseur<T> c) {
    Noeud n = (Noeud)c;
    n.filsAine = new Noeud(null, e, n.filsAine);
  }

  /**
   * Ajoute les noeuds de la fort <code>f</code> en tant que fils
   * du noeud dsign par <code>c</code>, et les supprime de <code>f</code>.
   * choue si ce noeud a dj des fils.
   */
  public void ajouterFille(Foret<T> f, Curseur<T> c) {
    Noeud n = (Noeud)c;

    if (n.filsAine != null) {
      throw new Error("Le noeud a dj une fille");
    }
    n.filsAine = f.racine_;
    f.racine_ = null; // car on a rcupr ses noeuds
  }

  /**
   * Ajoute un noeud contenant l'objet <code>e</code> en tant que
   * frre cadet (noeud situ immdaitement  droite) du noeud
   * dsign par <code>c</code>.
   */
  public void ajouterfrereCadet(T e, Curseur<T> c) {
    Noeud n = (Noeud)c;
    n.frereCadet = new Noeud(null, e, n.frereCadet);
  }

  /**
   * Ajoute les noeuds de la fort <code>f</code> en tant que frres
   * du noeud dsign par <code>c</code>, et les supprime de <code>f</code>.
   * choue si le noeud a dj des frres.
   */
  public void ajouterSoeur(Foret<T> f, Curseur<T> c) {
    Noeud n = (Noeud)c;

    if (n.frereCadet != null) {
      throw new Error("Le noeud a dj une soeur");
    }
    n.frereCadet = f.racine_;
    f.racine_ = null; // car on a rcupr ses noeuds
  }

  /**
   * Supprime les noeuds de la fort fille du noeud dsign
   * par <code>c</code>.
   */
  public void supprimerFille(Curseur<T> c) {
    ((Noeud)c).filsAine = null;
  }

  /**
   * Supprime les noeuds de la fort soeur du noeud
   * dsign par <code>c</code>.
   */
  public void supprimerSoeur(Curseur<T> c) {
    ((Noeud)c).frereCadet = null;
  }

  /**
   * Reprsente un noeud de la fort, et correspond  la notion
   * de curseur.
   */
  private class Noeud extends Curseur<T> {
    /** Objet contenu dans ce noeud. */
    public T contenu;

    /** Premier fils (le plus  gauche). */
    public Noeud filsAine;

    /** Frre cadet (situ immdiatement  droite). */
    public Noeud frereCadet;

    /** Construit un noeud contenant l'objet <code>e</code>. */
    public Noeud(T e) {
      this(null, e, null);
    }

    /**
     * Construit un noeud contenant l'objet <code>e</code>, ayant
     * pour fils ain <code>aine</code> et pour frre cadet <code>cadet</code>.
     */
    public Noeud(Noeud aine, T e, Noeud cadet) {
      contenu = e;
      filsAine = aine;
      frereCadet = cadet;
    }
  }
}
/*-*/