/* -*-Java-*-
 * ###################################################################
 *
 *  FILE: "ListeRecTableau.java"
 *                                    created: 2003-02-10 11:14:47
 *                                last update: 2008-02-27 10:05:33
 *  Author: Frdric Boulanger
 *  E-mail: Frederic.Boulanger@supelec.fr
 *    mail: Suplec - Service Informatique
 *          3 rue Joliot-Curie, F-91192 Gif-sur-Yvette cedex
 *     www: http://wwwsi.supelec.fr/fb/
 *
 *  Description:
 *  Liste rcursive dont les lments sont rangs dans un tableau.
 *
 *  History
 *
 *  modified   by  rev reason
 *  ---------- --- --- -----------
 *  2003-02-10 FBO 1.0 original
 * ###################################################################
 */
package collections;

import java.lang.reflect.Array;

/**
 * Liste rcursive dont les lments sont rangs dans un tableau.
 */
/*+ListeRecTableau/base.java*/
public class ListeRecTableau<T> extends ListeRec<T> {
  /** Tableau des lments. */
  private Conteneur[] elems_;

  /** Nombre d'lments dans la liste. */
  private int longueur_;

  // Oprations de la spcification.
  // Constructeur = listevide
  public ListeRecTableau() {
    elems_ = creerTableau(0);
    longueur_ = 0;
  }

  public void ajouterDebut(T e) {
    if (longueur_ + 1 > elems_.length) {
      Conteneur[] nouveau = creerTableau(longueur_+1);
      for (int i = 0; i < longueur_; i++) {
        nouveau[i] = elems_[i];
      }
      elems_ = nouveau;
    }
    elems_[longueur_] = new Conteneur(longueur_,e);
    longueur_++;
  }

  public void enleverPremier() {
    prerequis(longueur_ > 0);
    elems_[longueur_ - 1] = null;
    longueur_--;
  }

  public T premier() {
    prerequis(longueur_ > 0);
    return elems_[longueur_ - 1].element;
  }

  public boolean estvide() {
    return longueur_ == 0;
  }

  public int longueur() {
    return longueur_;
  }
/*-*/
/*+ListeRecTableau/creerTableau.java*/
  @SuppressWarnings("unchecked")
  private Conteneur[] creerTableau(int taille) {
    return (Conteneur[])Array.newInstance(Conteneur.class, taille);
  }
/*-*/  
  
  /**
   * Implmente la notion de curseur : comme les lments sont ici rangs
   * dans le tableau en ordre inverse de celui de la liste, on utilisera
   * le dcalage par rapport  la fin de la liste.
   */
/*+ListeRecTableau/conteneur.java*/
  private class Conteneur extends Curseur<T> {
    /**
     * On ne protge pas les membres de cette classe. En effet,
     * <code>Conteneur</code> est une classe prive qui ne peut tre
     * vue par aucune autre classe que <code>ListeRecTableau</code>.
     * De plus, son comportement est intimement li  celui de la liste.
     * Il est donc inutile de diminuer l'efficacit du code par des
     * appels de mthodes qui n'apportent rien au niveau de l'encapsulation
     * ou du polymorphisme.
     * Notez qu'il s'agit l d'une situation exceptionnelle.
     */
    public int idx;
    public T element;

    /** Construit un Curseur pointant sur l'lment <code>p</code> */
    public Conteneur(int p, T e) {
      idx = p;
      element = e;
    }
  }
/*-*/

  // Oprations sur les curseurs.
  //
  // Les oprations qui reoivent un Curseur devraient vrifier que
  // ce curseur appartient bien  la liste. Ici, cette vrification
  // n'est pas faite, et utiliser un curseur invalide provoquera donc
  // un comportement erratique sans qu'une erreur soit forcment signale.
  // Une solution consisterait  placer une rfrence sur la liste dans
  // laquelle le Curseur a un sens et  vrifier que cette liste est
  // la bonne  chaque appel. Cela viterait les comportements erratiques
  // silencieux mais dgraderait les performances.

/*+ListeRecTableau/methodesAvecCurseurs1.java*/
  public Curseur<T> debut() {
    if (longueur_ == 0) {
      return null;
    }
    return elems_[longueur_ - 1];
  }

  public Curseur<T> suivant(Curseur<T> c) {
    Conteneur ct = (Conteneur)c;

    if (ct.idx == 0) {
      return null;
    } else {
      return elems_[ct.idx - 1];
    }
  }

  public T element(Curseur<T> c) {
    Conteneur ct = (Conteneur)c;
    return elems_[ct.idx].element;
  }

  public void remplacer(Curseur<T> c, T e) {
    Conteneur ct = (Conteneur)c;
    elems_[ct.idx].element = e;
  }

  public void supprimerSuivant(Curseur<T> c) {
    Conteneur ct = (Conteneur)c;
    if (ct.idx == 0) {
      throw new Error("Curseur invalide pour supprimerSuivant");
    }
    for (int i = ct.idx; i < longueur_; i++) {
      elems_[i-1] = elems_[i];
      elems_[i-1].idx = i-1;
    }
    elems_[longueur_ - 1] = null;
    longueur_--;
  }
/*-*/
/*+ListeRecTableau/methodesAvecCurseurs2.java*/
  public void insererApres(Curseur<T> c, T e) {
    Conteneur ct = (Conteneur)c;
    int idx;

    if (ct == null)
      idx = longueur_;
    else
      idx = ct.idx;
    if (longueur_+1 > elems_.length) {
      Conteneur[] nouveau = creerTableau(longueur_+1);
      for (int i = longueur_; i > idx; i--) {
        nouveau[i] = elems_[i-1];
        nouveau[i].idx = i;
      }
      nouveau[idx] = new Conteneur(idx,e);
      for (int i = idx - 1; i >= 0; i--) {
        nouveau[i] = elems_[i];
      }
      elems_ = nouveau;
    } else {
      for (int i = longueur_; i > idx; i--) {
        elems_[i] = elems_[i-1];
        elems_[i].idx = i;
      }
      elems_[idx] = new Conteneur(idx,e);
    }
    longueur_++;
  }
/*-*/
  protected ListeRecTableau<T> makeInstance() {
    return new ListeRecTableau<T>();
  }
  
  // Mthode de test
  public static void main(String[] args) {
    ListeRec<Object> l = new ListeRecTableau<Object>();
    l.ajouterDebut("d");
    ListeRec<String> l2 = new ListeRecTableau<String>();
    l2.ajouterDebut("c");
    l2.ajouterDebut("b");
    l.ajouterDebut(l2);
    l.ajouterDebut("a");
    System.out.println(l.toString());
    l2.ajouterDebut("b'");
    l.ajouterDebut("a'");
    System.out.println(l.toString());
    l2.enleverPremier();
    l.enleverPremier();
    System.out.println(l.toString());
    Curseur<Object> c = l.debut();
    l.remplacer(c,"A");
    Curseur<String> c2 = l2.debut();
    l2.remplacer(c2,"B");
    System.out.println(l.toString());
    l.insererApres(c,"A'");
    System.out.println(l.toString());
    // Provoque une erreur car c n'est pas un curseur du bon type pour l2!
    // l2.insererApres(c,"Z");
    System.out.println(l.toString());
  }
}
